/*	$Id: snprintf.c,v 1.4 2005/08/22 21:48:35 phil Exp $ */

/* BLURB gpl

                           Coda File System
                              Release 6

          Copyright (c) 1987-2005 Carnegie Mellon University
                  Additional copyrights listed below

This  code  is  distributed "AS IS" without warranty of any kind under
the terms of the GNU General Public Licence Version 2, as shown in the
file  LICENSE.  The  technical and financial  contributors to Coda are
listed in the file CREDITS.

*/

#include <stdlib.h> /* abs */
#include <stdarg.h> /* va_list */

// Prototypes

static char *_xstrcpy(char *dst, char *src, char *end);
static char *_uinttostr(char *dst, unsigned int n, unsigned int base, char *end);
int snprintf(char *buf, size_t len, const char *fmt, ...);
int vsnprintf(char *buf, size_t len, const char *fmt, va_list ap);


static char *_xstrcpy(char *dst, char *src, char *end)
{
    char *p = dst;
    while(*src && p < end)
	*(p++) = *(src++);
    return p;
}

static char *_uinttostr(char *dst, unsigned int n, unsigned int base, char *end)
{
    char r, *p, tmp[33]; /* should be enough space for n as binary */

    if (base < 2 || base > 36)
	return dst;

    /* start at the end */
    p = &tmp[sizeof(tmp) - 1];
    *(p--) = '\0';

    if (!n) *(p--) = '0';

    while (n) {
	r = n % base;
	if (r < 10) *(p--) = '0' + r;
	else	    *(p--) = 'a' + r - 10;
	n /= base;
    }

    return _xstrcpy(dst, p+1, end);
}

int snprintf(char *buf, size_t len, const char *fmt, ...)
{
    va_list ap;
    int retval;
    va_start(ap, fmt);
    retval = vsnprintf(buf, len, fmt, ap);
    va_end(ap);
    return retval;
}

int vsnprintf(char *buf, size_t len, const char *fmt, va_list ap)
{
    int rc = 0;
    char *p = buf, *end;
    char *s;
    unsigned char c;
    unsigned int ui;
    int si;

    /* snprintf doesn't normally do this, but we make sure to leave room for
     * at least a terminating '\0' */
    len--;
    end = buf + len;

    for (; *fmt && p < end; fmt++) {
	if (*fmt != '%') {
	    *(p++) = *fmt;
	    continue;
	}
	fmt++;
	switch (*fmt) {
	    case '%':
		*(p++) = *fmt;
		break;

	    case 's':
		s = va_arg(ap, char *);
		p = _xstrcpy(p, s, end);
		break;

	    case 'c':
		c = (unsigned char)va_arg(ap, int);
		*(p++) = c;
		break;

	    case 'd':
	    case 'i':
		si = va_arg(ap, int);
		if (si < 0) *(p++) = '-';
		ui = abs(si);
		p = _uinttostr(p, ui, 10, end);
		break;

	    case 'o':
		ui = va_arg(ap, unsigned int);
		p = _uinttostr(p, ui, 8, end);
		break;

	    case 'u':
		ui = va_arg(ap, unsigned int);
		p = _uinttostr(p, ui, 10, end);
		break;

	    case 'x':
		ui = va_arg(ap, unsigned int);
		p = _uinttostr(p, ui, 16, end);
		break;

	    default:
		rc = -1;
	    case '\0':
		goto done;
	}
    }
done:
    if (rc) return rc;
    
    *p = '\0';
    return p - buf;
}

#ifdef TESTING
#include <stdio.h>
int main(int argc, char **argv)
{
    char buf[20];
    snprintf(buf, sizeof(buf), "012345678901234567890123456780"); puts(buf);
    snprintf(buf, sizeof(buf), "empty"); puts(buf);
    snprintf(buf, sizeof(buf), "%s", "const string"); puts(buf);
    snprintf(buf, sizeof(buf), "argv[0] = %s", argv[0]); puts(buf);
    snprintf(buf, sizeof(buf), "too long argv[0] %s", argv[0]); puts(buf);
    snprintf(buf, sizeof(buf), "argc = %d", argc); puts(buf);
    snprintf(buf, sizeof(buf), "-1971 = %d", -1971); puts(buf);
    snprintf(buf, sizeof(buf), "2004 = %d", 2004); puts(buf);
    snprintf(buf, sizeof(buf), "deadbeef = %x", 0xdeadbeef); puts(buf);
    snprintf(buf, sizeof(buf), "%d. %d. %d. %d. %s", 1, 2, 3, 4, "Five!"); puts(buf);
    return 0;
}
#endif
