/*	$Id: upcall.c,v 1.24 2005/08/25 20:22:51 phil Exp $ */

/* BLURB gpl

                           Coda File System
                              Release 6

          Copyright (c) 1987-2005 Carnegie Mellon University
                  Additional copyrights listed below

This  code  is  distributed "AS IS" without warranty of any kind under
the terms of the GNU General Public Licence Version 2, as shown in the
file  LICENSE.  The  technical and financial  contributors to Coda are
listed in the file CREDITS.

*/

/*
 * Mostly platform independent upcall operations to Venus:
 *  -- upcalls
 *  -- upcall routines
 *  -- downcall routine
 *
 * Linux 2.0 version
 * Copyright (C) 1996 Peter J. Braam <braam@maths.ox.ac.uk>, 
 * Michael Callahan <callahan@maths.ox.ac.uk> 
 * 
 * Redone for Linux 2.1
 * Copyright (C) 1997 Carnegie Mellon University
 *
 * Carnegie Mellon University encourages users of this code to contribute
 * improvements to the Coda project. Contact the coda team at coda@cs.cmu.edu.
 */

#define KERNEL 1

#include "errno.h"
#include <coda_types.h>
#include <coda.h>
#include <coda_nt.h>
#include <dllist.h>
#include <ipc.h>
#include <cnode.h>


#define CODA_ALLOC(buf, type, len) buf = (type) ExAllocatePool(NonPagedPool, len)

#define CODA_FREE(ptr,len)  ExFreePool(ptr)

#define UPARG(op, u_id)\
do {\
  	CODA_ALLOC(inp, union inputArgs *, insize);\
        if ( !inp ) { return -ENOMEM; }\
        outp = (union outputArgs *) (inp);\
        inp->ih.opcode = (op);\
	inp->ih.pid = 0;\
	inp->ih.pgid = 0;\
	inp->ih.uid = u_id;\
        outsize = insize;\
} while (0)


// static int max(int a, int b) 
//{
//    if ( a > b )
//	return a; 
//    else
//	return b;
//}

#ifndef max
#define max(a,b) (((a) < (b)) ? (b) : (a))
#endif

#define INSIZE(tag) sizeof(struct coda_ ## tag ## _in)
#define OUTSIZE(tag) sizeof(struct coda_ ## tag ## _out)
#define SIZE(tag)  max(INSIZE(tag), OUTSIZE(tag))

/* the upcalls */
int venus_rootfid(CodaFid *fidp)
{
    union inputArgs *inp;
    union outputArgs *outp;
    int insize, outsize, error;
    
    ENTRY(D_UPCALL, "venus_root");
    
    insize = SIZE(root);
    UPARG(CODA_ROOT,0);
    
    error = coda_upcall(insize, &outsize, inp);
    
    if (error) {
	CDEBUG((D_UPCALL,"coda_get_rootfid: error %d\n", error));
    } else {
	*fidp = outp->coda_root.Fid;
	CDEBUG((D_SUPER, "VolumeId: 0x%x 0x%x 0x%x 0x%x\n", FID(fidp)));

    }
    
    if (inp) CODA_FREE(inp, insize);
    
    EXIT(D_UPCALL, "venus_root");
    return error;
}

int venus_getattr(CodaFid *fid, struct coda_vattr *attr, int uid) 
{
    union inputArgs *inp;
    union outputArgs *outp;
    int insize, outsize, error;
    
    ENTRY(D_UPCALL, "venus_getattr");
    
    insize = SIZE(getattr); 
    UPARG(CODA_GETATTR, uid);
    inp->coda_getattr.Fid = *fid;
    
    error = coda_upcall(insize, &outsize, inp);
    
    if ( !error )
	*attr = outp->coda_getattr.attr;
    
    if (inp) CODA_FREE(inp, insize);
    
    EXIT(D_UPCALL, "venus_getattr");
    return error;
}

int  venus_setattr(CodaFid *fid, struct coda_vattr *attr, int uid)
{
    union inputArgs *inp;
    union outputArgs *outp;
    int insize, outsize, error;
    
    insize= SIZE(setattr);
    UPARG(CODA_SETATTR, uid);
    
    inp->coda_setattr.Fid = *fid;
    inp->coda_setattr.attr = *attr;

    CDEBUG ((D_UPCALL, "venus_setattr: %s atime = %d\n",
	     coda_f2s(fid), inp->coda_setattr.attr.va_atime.ts_sec));
    
    error = coda_upcall(insize, &outsize, inp);
    
    if ( inp ) CODA_FREE(inp, insize);
    
    CDEBUG ((D_UPCALL, "venus_setattr: returning error %d\n", error));
    return error;
}

int venus_lookup(CodaFid *fid, const char *name, int length,
		 int * type, CodaFid *resfid, int uid)
{
    union inputArgs *inp;
    union outputArgs *outp;
    int insize, outsize, error;
    int offset;
    
    ENTRY(D_UPCALL, "venus_lookup");
    
    offset = INSIZE(lookup);
    insize =  max(offset + length +1, OUTSIZE(lookup));
    UPARG(CODA_LOOKUP, uid);
    
    inp->coda_lookup.Fid = *fid;
    inp->coda_lookup.name = offset;
    inp->coda_lookup.flags = 1;
    
    /* send Venus a null terminated string */
    memcpy((char *)(inp) + offset, name, length);
    *((char *)inp + offset + length) = '\0';
    
    error = coda_upcall(insize, &outsize, inp);
    
    if ( !error ) {
	*resfid = outp->coda_lookup.Fid;
	*type = outp->coda_lookup.vtype;
    }

    if (inp) CODA_FREE(inp, insize);
    
    CDEBUG ((D_UPCALL, "venus_lookup: returning error %d\n", error));
    return error;
}


int venus_close(CodaFid *fid, int flags, int uid)
{
    union inputArgs *inp;
    union outputArgs *outp;
    int insize, outsize, error;
    
    insize = SIZE(close);
    UPARG(CODA_CLOSE, uid);
    
    inp->coda_close.Fid = *fid;
    inp->coda_close.flags = flags;
    
    CDEBUG ((D_UPCALL, "venus_close: fid = %s, flags = 0x%d \n",
	     coda_f2s(fid), flags));

    error = coda_upcall(insize, &outsize, inp);
    
    if (inp) CODA_FREE(inp, insize);
    
    CDEBUG((D_UPCALL, "venus_close: returning error %d\n", error));
    return error;
}


int venus_open_by_path(CodaFid *fid, int flags, char *cfname, int uid)
{
    union inputArgs *inp;
    union outputArgs *outp;
    int insize, outsize, error;
    
    insize = SIZE(open_by_path)+CODA_MAXPATHLEN;
    UPARG(CODA_OPEN_BY_PATH, uid);
    
    inp->coda_open_by_path.Fid = *fid;
    inp->coda_open_by_path.flags = flags;

    CDEBUG ((D_UPCALL, "venus_open_by_path: fid = %s, flags = 0x%x\n",
	     coda_f2s(fid), flags));
    
    error = coda_upcall(insize, &outsize, inp);
    
    if ( !error ) {
	strcpy(cfname, ((char *) outp) + 
	       outp->coda_open_by_path.path);
	CDEBUG ((D_UPCALL, "venus_open_by_path: file is %s\n", cfname));
    } 
    
    if (inp) CODA_FREE(inp, insize);
    
    CDEBUG ((D_UPCALL, "venus_open_by_path: returning error %d\n", error));
    return error;
}	


int venus_mkdir(CodaFid *dirfid, const char *name, int length, 
		CodaFid *newfid, struct coda_vattr *attrs, int uid)
{
    union inputArgs *inp;
    union outputArgs *outp;
    int insize, outsize, error;
    int offset;
    
    ENTRY(D_UPCALL, "venus_mkdir");
    
    offset = INSIZE(mkdir);
    insize = max(offset + length + 1, OUTSIZE(mkdir));
    UPARG(CODA_MKDIR, uid);
    
    inp->coda_mkdir.Fid = *dirfid;
    inp->coda_mkdir.attr = *attrs;
    inp->coda_mkdir.name = offset;
    
    /* Venus must get null terminated string */
    memcpy((char *)(inp) + offset, name, length);
    *((char *)inp + offset + length) = '\0';
    
    error = coda_upcall(insize, &outsize, inp);
    
    *attrs = outp->coda_mkdir.attr;
    *newfid = outp->coda_mkdir.Fid;
    
    if (inp) CODA_FREE(inp, insize);
    
    EXIT(D_UPCALL, "venus_mkdir");
    return error;        
}


int venus_rename(CodaFid *old_fid, const char *old_name, 
		 size_t old_length, 
		 CodaFid *new_fid, const char *new_name,
		 size_t new_length, int uid)
{
    union inputArgs *inp;
    union outputArgs *outp;
    int insize, outsize, error; 
    int offset, s;
    
    ENTRY(D_UPCALL, "venus_rename");
    
    offset = INSIZE(rename);
    insize = max(offset + new_length + old_length + 8,
		 OUTSIZE(rename)); 
    UPARG(CODA_RENAME, uid);
    
    inp->coda_rename.sourceFid = *old_fid;
    inp->coda_rename.destFid =  *new_fid;
    inp->coda_rename.srcname = offset;
    
    /* Venus must receive an null terminated string */
    s = ( old_length & ~0x3) +4; /* round up to word boundary */
    memcpy((char *)(inp) + offset, old_name, old_length);
    *((char *)inp + offset + old_length) = '\0';
    
    /* another null terminated string for Venus */
    offset += s;
    inp->coda_rename.destname = offset;
    s = ( new_length & ~0x3) +4; /* round up to word boundary */
    memcpy((char *)(inp) + offset, new_name, new_length);
    *((char *)inp + offset + new_length) = '\0';
    
    CDEBUG((D_UPCALL, "venus_rename:  destname in packet: %s\n", 
	   (char *)inp + (int) inp->coda_rename.destname));

    error = coda_upcall(insize, &outsize, inp);
    
    if (inp) CODA_FREE(inp, insize);
    
    EXIT(D_UPCALL, "venus_rename");
    return error;
}

int venus_create(CodaFid *dirfid, 
		 const char *name, int length, int excl, int mode, int rdev,
		 CodaFid *newfid, struct coda_vattr *attrs, int uid) 
{
    union inputArgs *inp;
    union outputArgs *outp;
    int insize, outsize, error;
    int offset;
    
    ENTRY(D_UPCALL, "venus_create");
    
    offset = INSIZE(create);
    insize = max(offset + length + 1, OUTSIZE(create));
    UPARG(CODA_CREATE, uid);
    
    inp->coda_create.Fid = *dirfid;
    inp->coda_create.attr.va_mode = (unsigned short)mode;
    inp->coda_create.attr.va_rdev = rdev;
    inp->coda_create.excl = excl;
    inp->coda_create.mode = mode;
    inp->coda_create.name = offset;
    
    /* Venus must get null terminated string */
    memcpy((char *)(inp) + offset, name, length);
    *((char *)inp + offset + length) = '\0';
    
    error = coda_upcall(insize, &outsize, inp);
    
    *attrs = outp->coda_create.attr;
    *newfid = outp->coda_create.Fid;
    
    if (inp) CODA_FREE(inp, insize);
    
    EXIT(D_UPCALL, "venus_create");
    return error;        
}

int venus_rmdir(CodaFid *dirfid, const char *name, int length, int uid)
{
    union inputArgs *inp;
    union outputArgs *outp;
    int insize, outsize, error;
    int offset;
    
    ENTRY(D_UPCALL, "venus_rmdir");
    
    offset = INSIZE(rmdir);
    insize = max(offset + length + 1, OUTSIZE(rmdir));
    UPARG(CODA_RMDIR, uid);
    
    inp->coda_rmdir.Fid = *dirfid;
    inp->coda_rmdir.name = offset;
    memcpy((char *)(inp) + offset, name, length);
    *((char *)inp + offset + length) = '\0';
    
    error = coda_upcall(insize, &outsize, inp);
    if ( inp ) 
	CODA_FREE(inp, insize);
    
    EXIT(D_UPCALL, "venus_rmdir");
    return error;
}

int venus_remove(CodaFid *dirfid, const char *name, int length, int uid)
{
    union inputArgs *inp;
    union outputArgs *outp;
    int error=0, insize, outsize, offset;
    
    ENTRY(D_UPCALL, "venus_remove");
    
    offset = INSIZE(remove);
    insize = max(offset + length + 1, OUTSIZE(remove));
    UPARG(CODA_REMOVE, uid);
    
    inp->coda_remove.Fid = *dirfid;
    inp->coda_remove.name = offset;
    memcpy((char *)(inp) + offset, name, length);
    *((char *)inp + offset + length) = '\0';
    
    error = coda_upcall(insize, &outsize, inp);
    if ( inp ) 
	CODA_FREE(inp, insize);
    
    EXIT(D_UPCALL, "venus_remove");
    return error;
}

#if 0
int venus_readlink(struct super_block *sb, CodaFid *fid, 
		   char *buffer, int *length, int uid)
{ 
    union inputArgs *inp;
    union outputArgs *outp;
    int insize, outsize, error;
    int retlen;
    char *result;
    
    ENTRY(D_UPCALL, "venus_readlink");
    
    insize = max(INSIZE(readlink), OUTSIZE(readlink)+ *length + 1);
    UPARG(CODA_READLINK);
    
    inp->coda_readlink.Fid = *fid;
    
    error =  coda_upcall(insize, &outsize, inp);
    
    if (! error) {
	retlen = outp->coda_readlink.count;
	if ( retlen > *length )
	    retlen = *length;
	*length = retlen;
	result =  (char *)outp + (int)outp->coda_readlink.data;
	memcpy(buffer, result, retlen);
	*(buffer + retlen) = '\0';
    }
    
    if (inp) CODA_FREE(inp, insize);

    CDEBUG((D_UPCALL, "venus_readlink: returning result %d\n",error));

    return error;
}

int venus_link(struct super_block *sb, CodaFid *fid, 
	       CodaFid *dirfid, const char *name, int len )
{
    union inputArgs *inp;
    union outputArgs *outp;
    int insize, outsize, error;
    int offset;
    
    ENTRY(D_UPCALL, "venus_link");
    
    offset = INSIZE(link);
    insize = max(offset  + len + 1, OUTSIZE(link));
    UPARG(CODA_LINK);
    
    inp->coda_link.sourceFid = *fid;
    inp->coda_link.destFid = *dirfid;
    inp->coda_link.tname = offset;
    
    /* make sure strings are null terminated */
    memcpy((char *)(inp) + offset, name, len);
    *((char *)inp + offset + len) = '\0';
    
    error = coda_upcall(insize, &outsize, inp);
    
    if (inp) CODA_FREE(inp, insize);

    CDEBUG((D_UPCALL, "venus_link: returning result %d\n",error));
    
    return error;
}

int venus_symlink(struct super_block *sb, CodaFid *fid,
		  const char *name, int len,
		  const char *symname, int symlen)
{
    union inputArgs *inp;
    union outputArgs *outp;
    int insize, outsize, error;
    int offset, s;
    
    ENTRY(D_UPCALL, "venus_symlink");
    
    offset = INSIZE(symlink);
    insize = max(offset + len + symlen + 8, OUTSIZE(symlink));
    UPARG(CODA_SYMLINK);
    
    /*        inp->coda_symlink.attr = *tva; XXXXXX */ 
    inp->coda_symlink.Fid = *fid;
    
    /* Round up to word boundary and null terminate */
    inp->coda_symlink.srcname = offset;
    s = ( symlen  & ~0x3 ) + 4; 
    memcpy((char *)(inp) + offset, symname, symlen);
    *((char *)inp + offset + symlen) = '\0';
    
    /* Round up to word boundary and null terminate */
    offset += s;
    inp->coda_symlink.tname = offset;
    s = (len & ~0x3) + 4;
    memcpy((char *)(inp) + offset, name, len);
    *((char *)inp + offset + len) = '\0';
    
    error = coda_upcall(insize, &outsize, inp);
    
    if (inp) CODA_FREE(inp, insize);

    CDEBUG((D_INODE, "venus_symlink: returning result %d\n",error));
    
    return error;
}

int venus_fsync(struct super_block *sb, CodaFid *fid)
{
    union inputArgs *inp;
    union outputArgs *outp; 
    int insize, outsize, error;
    
    ENTRY(D_UPCALL, "venus_fsync");
    
    insize=SIZE(fsync);
    UPARG(CODA_FSYNC);
    
    inp->coda_fsync.Fid = *fid;
    error = coda_upcall(sizeof(union inputArgs), 
			&outsize, inp);
    
    if ( inp ) 
	CODA_FREE(inp, insize);
    
    EXIT(D_UPCALL, "venus_fsync");
    return error;
}
#endif

int venus_access(CodaFid *fid, int flags, int uid)
{
    union inputArgs *inp;
    union outputArgs *outp; 
    int insize, outsize, error;
    
    
    ENTRY(D_UPCALL, "venus_access");
    
    insize = SIZE(access);
    UPARG(CODA_ACCESS, uid);
    
    inp->coda_access.Fid = *fid;
    inp->coda_access.flags = flags;
    
    error = coda_upcall(insize, &outsize, inp);
    
    if (inp) CODA_FREE(inp, insize);
    
    EXIT(D_UPCALL, "venus_access");
    return error;
}


//
// Statfs call gets information about the venus cache ...
//

int venus_statfs(struct coda_statfs *sfs) 
{ 
    union inputArgs *inp;
    union outputArgs *outp;
    int insize, outsize, error;
    
    ENTRY(D_UPCALL, "venus_statfs");
    insize = max(INSIZE(statfs), OUTSIZE(statfs));
    UPARG(CODA_STATFS, 0);
    
    error =  coda_upcall(insize, &outsize, inp);
    
    if (!error) {
	*sfs = outp->coda_statfs.stat;
    } else {
	CDEBUG ((D_UPCALL, "coda_statfs: Venus returns error %d\n", error));
    }
    
    if (inp) CODA_FREE(inp, insize);

    EXIT(D_UPCALL, "venus_statfs");
    return error;
}


//
// pioctl call ... mechanism for communicating with venus
//

int venus_pioctl(CodaFid *fid,
		 struct PioctlData* pdata, int pdatasize,
		 void *outbuf, int *outbufsize, int uid)
{
    union inputArgs *inp;
    union outputArgs *outp;  
    int insize, outsize, error;
    int iocsize;
    int cmd;
    
    CDEBUG ((D_UPCALL, "venus_pioctl: on %s\n", coda_f2s(fid)));
    
    insize = VC_MAXMSGSIZE;
    UPARG(CODA_IOCTL, uid);

    // Check sizes ...
    if (pdata->vi.in_size > VC_MAXDATASIZE) {
	DbgPrint ("venus_pioctl: bad input/output sizes\n");
	return STATUS_INVALID_DEVICE_REQUEST;
    }
    
    // build packet for Venus 
    
    inp->coda_ioctl.Fid = *fid;
    
    /* the cmd field was mutated by increasing its size field to
     * reflect the path and follow args. We need to subtract that
     * out before sending the command to Venus.  */

    cmd = pdata->cmd;

    inp->coda_ioctl.cmd = cmd;
    inp->coda_ioctl.len = pdata->vi.in_size;
    inp->coda_ioctl.data = (char *)(INSIZE(ioctl));

    /* get the input data data */
    if (pdata->vi.in_size > 0) 
	memcpy((char*)inp + (int)inp->coda_ioctl.data,
	       pdata->vi.in, pdata->vi.in_size);
	
    error = coda_upcall(INSIZE(ioctl) + pdata->vi.in_size, &outsize, inp);

    if (error) {
	CDEBUG((D_UPCALL,"coda_pioctl: Venus returns: %d for %s\n", 
	       error, coda_f2s(fid)));
	goto exit; 
    }
    
    /* Copy out the OUT buffer. */
    if (outp->coda_ioctl.len > pdata->vi.out_size
	|| outp->coda_ioctl.len > *outbufsize) {
	CDEBUG((D_UPCALL, "return len %d >= request len %d (bufsize %d)\n",
	       outp->coda_ioctl.len, pdata->vi.out_size, outbufsize));
	error = STATUS_INVALID_DEVICE_REQUEST;
    } else {
	RtlCopyMemory(outbuf, (char *)outp + (int)outp->coda_ioctl.data, 
		     MAX(pdata->vi.out_size,sizeof(int)));
	*outbufsize = pdata->vi.out_size;
    }
    
 exit:
    if (inp) CODA_FREE(inp, insize);
    
    EXIT(D_UPCALL, "venus_pioctl");
    return error;
}



/*  
    The statements below are part of the Coda opportunistic
    programming -- taken from the Mach/BSD kernel code for Coda. 
    You don't get correct semantics by stating what needs to be
    done without guaranteeing the invariants needed for it to happen.
    When will be have time to find out what exactly is going on?  (pjb)
*/


int coda_upcall(int insize, int *outsize, union inputArgs *inp)
{ 
    struct request *req;
    int rc;
    union outputArgs *outp;
    
    
    CDEBUG((D_IPC, "coda_upcall: Starting Call.\n"));
    
    req = request_allocate();
    if ( !req ) {
	CDEBUG( (D_UPCALL, "coda_upcall: request_allocate failed\n"));
	return STATUS_INSUFFICIENT_RESOURCES;
    }
    
    req->req_fetchlen = insize;
    req->req_answerlen = *outsize;
    req->req_bufflen = MAX(insize, *outsize);
    req->req_data.in = inp;
    
    CDEBUG((D_IPC, "coda_upcall: About to process_rpc\n"));
    
    rc = request_process(req);
    
    CDEBUG((D_IPC, "coda_upcall: process_rpc returned %d\n", rc));
    
    outp = (union outputArgs *)inp;
    if ( rc == 0 ) {
	rc = outp->oh.result;
    }
    
    request_free (req);
    
    CDEBUG((D_IPC, "coda_upcall: Returning.\n"));
    return rc;
}


// Not the best place for this .... but it is part of the venus->kernel
// communication.

/* 
 * There are 7 cases where cache invalidations occur.  The semantics
 *  of each is listed here:
 *
 * CODA_FLUSH     -- flush all entries from the name cache and the cnode cache.
 * CODA_PURGEUSER -- flush all entries from the name cache for a specific user
 *                  This call is a result of token expiration.
 *
 * The next arise as the result of callbacks on a file or directory.
 * CODA_ZAPFILE   -- flush the cached attributes for a file.
 
 * CODA_ZAPDIR    -- flush the attributes for the dir and
 *                  force a new lookup for all the children
 of this dir.
 
 *
 * The next is a result of Venus detecting an inconsistent file.
 * CODA_PURGEFID  -- flush the attribute for the file
 *                  purge it and its children from the dcache
 *
 * The last  allows Venus to replace local fids with global ones
 * during reintegration.
 *
 * CODA_REPLACE -- replace one CodaFid with another throughout the name cache */
NTSTATUS coda_downcall(IN int opcode, IN union outputArgs * out)
{
    
    switch (opcode) {
	
    case CODA_FLUSH : {
	CDEBUG((D_DOWNCALL,"coda_downcall: Flush\n"));
	break;
    }
    
    case CODA_PURGEUSER : {
	//	struct coda_cred *cred = &out->coda_purgeuser.cred;
	CDEBUG((D_DOWNCALL,"coda_downcall: PurgeUser\n"));
	//	if ( !cred ) {
	//	    CDEBUG((D_UPCALL, "     PURGEUSER: null cred!\n"));
	//	    return 0;
	//	}
	// coda_cache_clear_cred(cred);
	break;
    }

    //
    // Should actually remove the fid from the cache.
    //
    case CODA_PURGEFID : {
	coda_inode *inode;
	coda_filehandle *fh;
	CodaFid *fid = &out->coda_purgefid.Fid;

	CDEBUG((D_DOWNCALL, "coda_downcall: purgefid fid = %s...\n",
		coda_f2s(fid)));

	if (IS_CTL_FID(fid)) {
	    CDEBUG((D_DOWNCALL, "coda_downcall: not purging CTL fid\n"));
	    return STATUS_SUCCESS;
	}
	
	inode = inode_find_fid(fid);

	if (inode) {
	    CDEBUG ((D_DOWNCALL,
		     "coda_downcall: PurgeFID: inode flags=0x%x fh_list=%x\n",
		     inode->c_status, inode->c_fh_list));
	    inode_unsetflag(inode, I_VATTR);
	    inode_setflag(inode, I_PURGE);
	    if (CODA_ISDIR(inode)) {
		// coda_purge_children(inode); 
	    }
	    GrabLock();
	    fh = inode->c_fh_list;
	    if (fh) {
		CDEBUG ((D_DOWNCALL | D_DETAILS, "coda_downcall: purge fh\n"));
		DropLock();
		while (fh) {
		    (void) OwPurgeCache(fh);
		    fh = fh->fh_next;
		}
		CDEBUG ((D_DOWNCALL | D_DETAILS,
			 "coda_downcall: end purge fh\n"));
	    } else {
		inode_release(inode);
		DropLock();
	    }

	    CDEBUG ((D_DOWNCALL,
		     "coda_downcall: Purge End: inode flags=0x%x fh_list=%x\n",
		     inode->c_status, inode->c_fh_list));
	} else {
	    CDEBUG((D_DOWNCALL, "coda_downcall: purgefid: no inode\n"));
	}
	
	break;
    }

    //
    // Should only remove attributes of file
    //
    case CODA_ZAPDIR :
    case CODA_ZAPFILE : {
	coda_inode *inode;
	coda_filehandle *fh;
	CodaFid *fid = &out->coda_zapfile.Fid;
	CDEBUG((D_DOWNCALL, "coda_downcall: zapfid fid = %s...\n",
		coda_f2s(fid)));
	
	inode = inode_find_fid(fid);
	if (inode) {
	    inode_unsetflag(inode, I_VATTR);
	    CDEBUG((D_DOWNCALL, "coda_downcall: zapfid purging file handles\n"));
	    if (CODA_ISDIR(inode)) {
		if (inode_hasflag(inode, I_LOCKED)) {
		    CDEBUG((D_DOWNCALL, "coda_downcall: zapfid on locked inode\n"));
		    inode_setflag(inode, I_ZAPPED);
		} else {
		    fh = inode->c_fh_list;
		    while (fh != NULL) {
			CDEBUG((D_DOWNCALL, "coda_downcall: purging fh %d\n", fh->fh_unique));
			(void) OwPurgeCache(fh);  // XXX should use ret val
			fh = fh->fh_next;
		    }
		}
	    }
	} else {
	    CDEBUG((D_DOWNCALL, "coda_downcall: zapfid: no inode\n"));
	}
	
	CDEBUG((D_DOWNCALL, "coda_downcall: zapfid: done\n"));
	break;
    }

    
    case CODA_REPLACE : {
	struct coda_inode *inode;
	CodaFid *oldfid = &out->coda_replace.OldFid;
	CodaFid *newfid = &out->coda_replace.NewFid;
	CDEBUG((D_DOWNCALL, "coda_downcall: replace\n"));
#if 0
	inode = coda_fid_to_inode(newfid, sb);
	if ( inode ) { 
	    CDEBUG((D_DOWNCALL, "replacefid: inode = %ld\n", inode->i_ino));
	    coda_replace_fid(inode, oldfid, newfid);
	    coda_purge_dentries(inode);
	} else {
	    CDEBUG((D_DOWNCALL, "purgefid: no inode\n"));
	}
#endif
	
	break;
    }

    default:
	CDEBUG((D_DOWNCALL, "coda_downcall: unhandled opcode %d\n", opcode));
    }
    return STATUS_SUCCESS;
}

